import aiohttp
import logging
import uuid

from datetime import datetime, timedelta
from urllib.parse import urlencode, urlsplit, parse_qs

from .exceptions import UplinkResponseException, UplinkException


_LOGGER = logging.getLogger(__name__)


async def raise_for_status(response):
    if 400 <= response.status:
        e = aiohttp.ClientResponseError(
            response.request_info,
            response.history,
            code=response.status,
            headers=response.headers,
        )

        if "json" in response.headers.get("CONTENT-TYPE", ""):
            data = await response.json()
            e.message = str(data)
            raise UplinkResponseException(data.get("errorCode"), data) from e

        else:
            data = await response.text()
            raise UplinkException(data) from e


class BearerAuth(aiohttp.BasicAuth):
    def __init__(self, access_token):
        self.access_token = access_token

    def encode(self):
        return "Bearer {}".format(self.access_token)


class UplinkSession:
    def __init__(
        self,
        client_id,
        client_secret,
        redirect_uri,
        access_data=None,
        access_data_write=None,
        base="https://api.nibeuplink.com",
        scope=["READSYSTEM"],
    ):
        self.redirect_uri = redirect_uri
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_data_write = access_data_write
        self.access_data = None
        self.session = None
        self.scope = scope
        self.base = base

        # check that the access scope is enough, otherwise ignore
        if access_data:
            if set(scope).issubset(set(access_data["scope"].split(" "))):
                self.access_data = access_data
            else:
                _LOGGER.info(
                    "Ignoring access data due to changed scope {}".format(scope)
                )

    async def __aenter__(self):
        await self.open()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def open(self):

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded;charset=UTF-8",
        }

        self.session = aiohttp.ClientSession(
            headers=headers, auth=aiohttp.BasicAuth(self.client_id, self.client_secret)
        )

        if self.access_data:
            await self.refresh_access_token()

    async def close(self):
        if self.session:
            await self.session.close()
            self.session = None

    def _handle_access_token(self, data):
        if "access_token" not in data:
            raise ValueError("Error in reply {}".format(data))

        if "expires_in" in data:
            _LOGGER.debug("Token will expire in %s seconds", data["expires_in"])
            expires = datetime.now() + timedelta(seconds=data["expires_in"])
        else:
            expires = None

        data["access_token_expires"] = expires.isoformat()

        self.access_data = data
        if self.access_data_write:
            self.access_data_write(data)

    async def _get_auth(self):
        if self.access_data:
            return BearerAuth(self.access_data["access_token"])
        else:
            return None

    async def get_access_token(self, code):
        payload = {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": self.redirect_uri,
        }

        async with self.session.post(
            "{}/oauth/token".format(self.base), data=payload
        ) as response:
            await raise_for_status(response)
            self._handle_access_token(await response.json())

    async def refresh_access_token(self):
        if not self.access_data or "refresh_token" not in self.access_data:
            _LOGGER.warning("No refresh token available for refresh")
            return

        _LOGGER.debug(
            "Refreshing access token with refresh token %s",
            self.access_data["refresh_token"],
        )
        payload = {
            "grant_type": "refresh_token",
            "refresh_token": self.access_data["refresh_token"],
        }

        async with self.session.post(
            "{}/oauth/token".format(self.base), data=payload
        ) as response:
            await raise_for_status(response)
            self._handle_access_token(await response.json())

    def get_authorize_url(self, state=None):
        if not state:
            state = uuid.uuid4().hex
        self.state = state

        params = {
            "response_type": "code",
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "scope": " ".join(self.scope),
            "state": state,
        }

        return "{}/oauth/authorize?{}".format(self.base, urlencode(params))

    def get_code_from_url(self, url):
        query = parse_qs(urlsplit(url).query)
        if "state" in query and query["state"][0] != self.state:
            raise ValueError(
                "Invalid state in url {} expected {}".format(query["state"], self.state)
            )
        return query["code"][0]

    async def request(self, *args, **kw):
        response = await self.session.request(*args, auth=await self._get_auth(), **kw)
        try:
            if response.status == 401:
                _LOGGER.debug(response)
                _LOGGER.info("Attempting to refresh token due to error in request")
                await self.refresh_access_token()
                response.close()
                response = await self.session.request(
                    *args, auth=await self._get_auth(), **kw
                )

            await raise_for_status(response)

            if "json" in response.headers.get("CONTENT-TYPE", ""):
                data = await response.json()
            else:
                data = await response.text()

            return data

        finally:
            response.close()



"""Exceptions used for nibeuplink"""
from typing import List, Dict, Any

ERROR_CODES = {
    -2: ("NO_CHANGE", "Parameter value is the same as the current value."),
    -1: ("NO_ERROR", "Not an error, indicates success."),
    0: ("UNKNOWN_ERROR", "Unknown error occured."),
    1: ("FORMAT_ERROR", "Input format error."),
    2: ("UNKNOWN_SYSTEM", "Unknown system."),
    3: ("ALREADY_CONNECTED", "User is already connected to the system."),
    4: (
        "NOT_ALLOWED_AS_SERVICE_USER",
        "A service user is trying to do something he/she is not allowed to.",
    ),
    5: ("NEEDS_RECONNECT", "System needs a reconnect."),
    6: (
        "NOT_ALLOWED_TO_RECONNECT",
        "System is not allowed to be reconnected by this particular user.",
    ),
    7: ("NOT_FOUND", "The searched resource was not found."),
    8: ("NOT_POSSIBLE", "Could not accomplish the requested action."),
    9: ("UNKNOWN_SYSTEM_UNIT", "System unit requested was not found."),
    10: ("UNKNOWN_HEAT_SYSTEM", "Heat system was not found."),
    11: ("NOT_ALLOWED_ON_SYSTEM", "User is not allowed to do task on selected system."),
    12: (
        "NO_PREMIUM_ON_SYSTEM",
        "The system lacks premium subscriptions needed to access this function.",
    ),
    13: ("TOO_MANY", "Too many objects was requested."),
    14: ("INVALID", "Invalid."),
    15: ("OUT_OF_RANGE", "Parameter value is out of range."),
    16: ("NOT_SETABLE", "Parameter is not setable."),
    17: ("NO_METADATA", "No metadata for parameter id found."),
    18: ("UNKNOWN_PARAMETER", "Unknown parameter id."),
    19: ("OAUTH_AUTHORIZATION_FAILED", "OAuth Authorization Failed."),
    20: ("OAUTH_INVALID_SCOPE", "OAuth, insufficient scopes for resource."),
    21: (
        "OAUTH_INVALID_ROLES",
        "OAuth, user is missing role required to access resource.",
    ),
    22: (
        "OAUTH_NOT_ALLOWED_CLIENT",
        "OAuth, client is not allowed to access resource.",
    ),
    23: ("UNKNOWN_CLIENT", "Client connecting is unknown."),
    24: ("INVALID_CLIENT_GROUP", "Invalid client group."),
    25: ("NOT_ALLOWED_AS_DEMO_USER", "Not allowed as demo user."),
    26: ("SYSTEM_OFFLINE", "The system cannot be reached."),
    27: ("INVALID_VOUCHER_TYPE", "Invalid voucher type."),
    28: ("RATE_LIMIT", "Rate limit. Too many requests."),
    29: ("NOT_AVAILABLE", "Resource not available for user."),
    30: ("INVALID_VOUCHER_SERIAL_NUMBER", "Serial number not valid for voucher."),
    31: ("CANT_BE_REGISTERED", "Resource can't be registered."),
    32: ("REGISTRATION_UNAVAILABLE", "Registration is unavailable."),
}


class UplinkException(Exception):
    pass


class UplinkResponseException(UplinkException):
    def __init__(self, code: int, data: Dict[str, Any]):
        self.code = code
        self.data = data
        desc = ERROR_CODES.get(code)
        if desc:
            self.message = desc[1]
            self.name = desc[0]
        else:
            self.message = "Unknown error {}".format(code)
            self.name = "UNKNOWN_CODE"
        super().__init__("{}: {}\n{}".format(self.name, self.message, self.data))



"""Structures representing data"""
from typing import (
    Optional,
    List,
)

import attr

from .typing import ParameterId


@attr.s
class Parameter(object):
    parameterId = attr.ib()
    name = attr.ib()
    title = attr.ib()
    designation = attr.ib()
    unit = attr.ib()
    displayValue = attr.ib()
    rawValue = attr.ib()


@attr.s
class ParameterExtended(Parameter):
    value = attr.ib()


@attr.s(auto_attribs=True)
class ClimateSystem:
    name: str
    return_temp: Optional[ParameterId]  # BT3
    supply_temp: Optional[ParameterId]  # BT2
    calc_supply_temp_heat: Optional[ParameterId]  # CSTH
    offset_cool: Optional[ParameterId]  # OC
    room_temp: Optional[ParameterId]  # BT50
    room_setpoint_heat: Optional[ParameterId]  # RSH
    room_setpoint_cool: Optional[ParameterId]  # RSC
    use_room_sensor: Optional[ParameterId] # URS
    active_accessory: Optional[ParameterId]  # AA
    external_adjustment_active: Optional[ParameterId]  # EAA
    calc_supply_temp_cool: Optional[ParameterId]  # CSTC
    offset_heat: Optional[ParameterId]  # OH
    heat_curve: Optional[ParameterId]  # HC
    min_supply: Optional[ParameterId]  # MIS
    max_supply: Optional[ParameterId]  # MAS
    extra_heat_pump: Optional[ParameterId]  # EHP


@attr.s(auto_attribs=True)
class HotWaterSystem:
    name: str
    hot_water_charging: Optional[ParameterId]  # BT6
    hot_water_top: Optional[ParameterId]  # BT7
    hot_water_comfort_mode: Optional[ParameterId]
    hot_water_production: Optional[ParameterId]  # Active or not
    periodic_hot_water: Optional[ParameterId]
    stop_temperature_water_normal: Optional[ParameterId]
    start_temperature_water_normal: Optional[ParameterId]
    stop_temperature_water_luxary: Optional[ParameterId]
    start_temperature_water_luxary: Optional[ParameterId]
    stop_temperature_water_economy: Optional[ParameterId]
    start_temperature_water_economy: Optional[ParameterId]
    total_hot_water_compressor_time: Optional[ParameterId]
    hot_water_boost: Optional[ParameterId]


@attr.s(auto_attribs=True)
class VentilationSystem:
    name: str
    fan_speed: Optional[ParameterId]
    exhaust_air: Optional[ParameterId]  # BT20
    extract_air: Optional[ParameterId]  # BT21
    exhaust_speed_normal: Optional[ParameterId]
    exhaust_speed_1: Optional[ParameterId]
    exhaust_speed_2: Optional[ParameterId]
    exhaust_speed_3: Optional[ParameterId]
    exhaust_speed_4: Optional[ParameterId]
    ventilation_boost: Optional[ParameterId]
